from joblib import dump, load
import os
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    TextIteratorStreamer,
    GenerationConfig,
)
import json
from sentence_transformers import SentenceTransformer
import numpy as np


def save_model(model, model_path):
    dump(model, model_path)
    return model_path

def load_model(model_path):
    return load(model_path)

def save_tokenizer(tokenizer, tokenizer_path):
    dump(tokenizer, tokenizer_path)
    return tokenizer_path

def load_tokenizer(tokenizer_path):
    return load(tokenizer_path)

def load_chat_history(file_path):
    try:
        with open(file_path, 'r') as file:
            return json.load(file)
    except FileNotFoundError:
        return []  # Return an empty list if the file does not exist

def save_chat_history(chat_history, file_path):
    with open(file_path, 'w') as file:
        json.dump(chat_history, file, indent=4)

def update_system_message(chat_history, new_system_content):
    # Find the last 'system' message in the chat history
    system_message_found = False
    for message in reversed(chat_history):
        if message['role'] == 'system':
            message['content'] = new_system_content  # Update the content of the last 'system' message
            system_message_found = True
            break
    if not system_message_found:
        chat_history.append({"role": "system", "content": new_system_content})  # Add a new 'system' message if none exists

def relevant_history(chat_history, question):
    # Initialize the sentence transformer model for embeddings
    embedder = SentenceTransformer('all-MiniLM-L6-v2')

    # Function to preprocess and return embeddings for given texts
    def get_embeddings(texts):
        return embedder.encode(texts)

    system_prompt = chat_history[0]
    # Assuming pairs are structured as consecutive user and assistant messages
    chat_pairs = chat_history[1:]

    # Prepare texts for embedding
    # Concatenate user and assistant messages for pairs to keep their context
    pair_texts = [chat_pairs[i]['content'] + " " + chat_pairs[i+1]['content']
                  for i in range(0, len(chat_pairs), 2)]

    # Get embeddings for each pair and the query
    pair_embeddings = get_embeddings(pair_texts)
    query_embedding = get_embeddings([question])[0]

    # Compute cosine similarity between query and each pair
    cos_similarities = np.dot(pair_embeddings, query_embedding) / (
        np.linalg.norm(pair_embeddings, axis=1) * np.linalg.norm(query_embedding)
    )

    # Get indices of the top 3 most similar pairs
    top_indices = np.argsort(cos_similarities)[-3:][::-1]

    # Re-rank the chat pairs based on these indices
    reranked_pairs = []
    for index in top_indices:
        reranked_pairs.extend(chat_pairs[index*2:index*2+2])  # Each pair has two messages

    # Reintegrate the system prompt at the beginning
    reranked_chat = [system_prompt] + reranked_pairs

    return reranked_chat